#ifndef __MAGIC_CSTRING_H__
#define __MAGIC_CSTRING_H__

#include <magic/object.h>
#include <string.h> // fgetS ja safedup vaativat
#include <stdlib.h> // atoi()

// Externals
class RegExp;



///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//                         ----         o                                    //
//                        (      |          _                                //
//                         ---  -+- |/\ | |/ \   ___                         //
//                            )  |  |   | |   | (   \                        //
//                        ___/    \ |   | |   |  ---/                        //
//                                               __/                         //
///////////////////////////////////////////////////////////////////////////////

enum quoteflags {QUOTE_NORMAL=0, QUOTE_HTML=1};

/** Generic string and buffer class. It supports cached hash-number
 *  calculation.
**/
class String : public Comparable {
	decl_dynamic (String);
	
	int				maxlen;	// Varattu pituus
	char*			data;
	unsigned char	chksum;
  public:

	/** Current length of the string. This *SHOULD* be a function,
	 *  but...
	 **/
	int		len;

				String			();
				String			(const String& orig);

	/** NULL-terminated string. */
				String			(const char*);

	/** NULL-terminated string with given length. */
				String			(const char*, int n);

	/** Conversion from an integer */
				String			(int i);
	/** Conversion from a float */
				String			(float f);
	/** Conversion from a double */
				String			(double f);

	virtual		~String		();

 public:
	// Implementations

	/**
	   Defined for the use of Comparable::hashfunc.  The hash
	   function is simply h(c)=c%hashsize. Once the hash number
	   has been calculated for the string, it is cached. This
	   gives a considerable speedup for evaluating this method.
	*/
	int			hashfunc		(int hashsize) const;
	
	// Common operators

	/** Assignment from a single character. */
	String&		assign			(char c);

	/** Assignment from another String */
	String&		operator=		(const String&);

	/** Assignment from a C-string. */
	String&		operator=		(const char*);

	/** Assignment from a single character. */
	String&		operator=		(char c) {return assign(c);}

	/** Not equals */
	int			operator!=		(const char* b) const {return !operator== (b);}

	/** Equals */
	int			operator==		(const char*) const;

	/** Equals */
	int			operator==		(const String& b) const {return (*this) == b.data;}

	/**
       Checks if the String equals another Comparable, which *MUST*
	   be a string. This implements the Comparable::operator== that is
	   used for sorting
	*/
	int			operator==		(const Comparable& other) const;

	/** Concatenates another String to the string */
	String&		operator+=		(const String& str);

	/** Concatenates a single character to the string */
	String&		operator+=		(char c);

	/** Concatenates a String to the string and returns the result */
	String		operator+		(const String& str) const;

	/** Concatenates a C-string to the string and returns the result */
	String		operator+		(const char* str) const;

	/** Returns a const reference to the n:th character in the string */
	const char&	operator[]		(int n) const {return (n>=0 && n<len)? data [n] : nullchar;}

	/** Returns a non-const reference to the n:th character in the string */
	char&		operator[]		(int n) {return (n>=0 && n<len)? data [n] : nullchar;}

	// Manipulations
	
	/**
	   Splits the string into an array according to the given
	   delimiter.
	*/
	void		split			(Array<String>& target, const char delim=' ') const;

	/**
	   Forms the string by joining the substrings in the given array
	   with the given delimiter.
	*/
	void		join			(const Array<String>& source, const char delim=' ');

	/** Returns a substring (0-based indexing). */
	String		substr			(int start, int len=-1) const;
	String		left			(int n) const;
	String		right			(int n) const;

	/** Deletes the last n characters for the string */
	String&		dellast			(int n);

	/** Removes all trailing newline (\r, \n) characters. */
	void		chop			();

	/**
	   Returns the string without any whitespace (space, tab) 
	   characters in the beginning or the end.
	*/
	String		trim			() const;

	void		uppercase		() const;
	void		lowercase		() const;

	/** Searches for a substring. Returns -1 of not found. */
	int			find			(const String&, int start=0) const;

	/** Searches for a single character. Returns -1 of not found. */
	int			find			(const char c, int start=0) const;

	/**
	   Reverse substring search. If start<0, it is calculated from the
	   end of the string. Returns -1 if not found.
	*/
	int			rfind			(const String&, int start=-1) const;
	
	/**
	   Tries to match the given regular expression to the
	   string. Returns != 0 if succesful.
	*/
	int			regmatch		(const char* regexpr) const;

	/**
	   The second version supports parameters in the regular
	   expression.  Returns != 0 of succesful and places the return
	   values into the array.
	*/
	int			regmatch		(const char* regexpr, Array<String>& target) const;

	/** As above, but uses a precompiled regular expression. */
	int			regmatch		(RegExp& compiled, Array<String>& target) const;

	/** Is the string NULL? */
	int			isnull			() const {if (this) return !data; return 0;}

	/** Is the string empty (can be empty without being NULL)? */
	int			isempty			() const {return (!this || !data || !len);}

	/** Ensures that at least the given length is reserved for the string. */
	void		ensure			(int amount) {if (maxlen<amount) reserve (amount);}

	/** Reserves exactly the given length for the string (shortens the string if necessary). */
	void		reserve			(int amount);

	/**
	   Ensures that at least the given amount is reserved for the
	   string, preferably a little (50%+4 characters) more.
	*/
	void		ensure_spontane	(int amount) {if (maxlen<amount) reserve (amount+amount/2+4);}

	/** Grows the string somewhat (50%+4 characters) */
	void		grow_spontane	() {reserve (maxlen+maxlen/2+4);}

	/** Calculates an 8-bit checksum for the string */
	char		checksum		();

	/**
	   Fast comparison operator, based on cached checksum (hash
	   value). Very quick if the hash values differ. OBSERVE! This
	   method doesn't calculate the checksums for the strings; you
	   must calculate them manually!
	*/
	int			fast_isequal	(const String& other) const {
		if (other.chksum == chksum)
			return operator== (other.data);
		else
			return 0;
	}

	/** Returns the current maximum length of the buffer.
	 **/
	int			maxLen			() const {return maxlen;}

	// Encodings

	// Codes the given string in two-character hex code.
	String&		hexcode			(const String& other);

	// Quote or unquote non-printable characters with given escape
	// character. QUOTE_HTML-flag quotes the spaces into
	// +-characters.
	void		quote			(char quotechar='%', int flags=0);
	void		unquote			(char quotechar='%', int flags=0);


	// I/O
	
	ostream&	dump			(ostream&) const;
	ostream&	operator>>		(ostream&) const;
	istream&	operator<<		(istream&);
	CArchive&	operator>>		(CArchive&) const;
	IStream&	operator<<		(IStream&);
	Object*		clone			() const;

	// Conversions
	
				operator const char* () const {return this? data : (const char*) NULL;}
				operator int		() const {if (this && data) return atoi (data); return 0;}
				operator long		() const {if (this && data) return atol (data); return 0;}
				operator float		() const;
				operator double		() const {if (data) return atof (data); return 0;}

	// Returns a non-const pointer to the string buffer. Dangerous.
	char*		getbuffer		() const {return data;}

	friend String	strformat	(const char* format, ...);
};

#define STR		char*

// Tyhj stringi, tm voidaan palauttaa tarvittaessa JA SE TULEE AINA
// PIT TYHJN PERKELE. Huoh, aika toivoton pyrkimys kyllkin...

extern String emptystring;

inline bool isempty (const String& str) {
	if (!(&str)) // isnull?
		return true;
	if (str.isempty ())
		return true;
	return false;
}

// Hakee n-pituisesta merkkijonosta str merkin c,
// palauttaa osoittimen paikkaan tai NULL jos ei lydy
char*	strnchr	(const char* str, int len, char c);

// Safer version of strdup - checks if the string is NULL.
inline char* safedup (const char* orig) {
	if (orig) {
		// Duplicate.
		// We can't use strdup, because ANSI C++ doesn't have it.
		int len = strlen(orig)+1; // Length with end-null.
		char* result = new char [len];
		memcpy (result, orig, len);
		return result;
	} else
		return (char*) NULL;
}

// Lukee yhden \n-terminoidun rivin streamista String-merkkijonoon.
// Palauttaa luettujen merkkien mrn (-1 jos EOF).
int			fgetS	(FILE* in, String& str);

// Reads a string from the stream. Terminated by given character, or
// with whitespace if -1.
istream&	getS	(istream& in, String& str, char term='\xff');

// Loads string from file
void loadString (String& str, const String& filename);

/**
   <H3>const CString& <B>format</B> (const char* <I>format</I>, ...)</H3>
   - Muotoilee parametrit annetun (printf-) formaatin mukaan ja palauttaa
   - tuloksen merkkijonona.<P>
   - <B>HUOM!</B> Maksimikoko 1024 merkki. Jos tm ylittyy, keskeytetn
   - ohjelman suoritus vlittmsti, koska muistia on mit ilmeisimmin
   - sotkeentunut.
*/
String	strformat	(const char* format, ...);
#ifndef format
#define format strformat
#endif

#define CONSTR	const char*


///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//                ----              ----         o                           //
//               (           |     (      |          _                       //
//                ---  |   | |---   ---  -+- |/\ | |/ \   ___                //
//                   ) |   | |   )     )  |  |   | |   | (   \               //
//               ___/   \__! |__/  ___/    \ |   | |   |  ---/               //
//                                                        __/                //
///////////////////////////////////////////////////////////////////////////////

// Virtual substring of a string
class SubString : public String {
	String*		str;
	int			start, end;
  public:
};

#endif
