///////////////////////////////////////////////////////////////////////////////  
//  Copyright Marko Grnroos 1996, all rights reserved.                     //
// Omistusoikeus Marko Grnroos, 1996, kaikki oikeudet pidtetn.           //
//   Tmn tiedoston kopiointi, lukeminen ja kytt ja erityisesti tmn     //
// ilmoitustekstin muuttaminen on kielletty ilman asianmukaista lupaa.      //
//                                                                           //
//                                             Marko Grnroos (magi@utu.fi)  //
///////////////////////////////////////////////////////////////////////////////

#ifndef __MAGISUPP_H__
#define __MAGISUPP_H__


#include "magic/mconfig.h"

// Remove new-redefinition
#ifdef new
#undef new
#define new_UNDEFD
#endif

#include<stdexcept>

// Restore new-redefinition
#ifdef new_UNDEFD
#define new DEBUG_NEW
#undef new_UNDEFD
#endif

#include "magic/debug.h"

class String;

///////////////////////////////////////////////////////////////////////////////
// Exception handling macros

/** Highest exception class. Replacement for standard C++ 'exception'
 *  class. I think there was some reason to replace it.
 *
 *  You can subclass this baseclass with the EXCEPTIONCLASS()
 *  macro. For example:
 *
 *  EXCEPTIONCLASS (assertion_failed);
 *
 *  See magisupp.h for some common exception classes. Many debugging
 *  macros that throw exceptions are also defined there.
 *
 *  Exception fails can be tracked with the FAILTRACE macro.
 **/
class Exception : public Object {
	String*	mWhat;
  public:

	/** Constructor the Exception with the given explanation.
	 *
	 *  Example:
	 *
	 *  throw Exception ("Some text");
	 **/
					Exception	(const String& what);
					Exception	(const Exception& orig);
					~Exception	();

	/** Returns the explanation text given at throw.
	 **/
	const String&	what		() const {return *mWhat;}
};

#define EXCEPTIONCLASS(eclass) \
class eclass : public Exception {\
public:\
	eclass (const String& what_arg) : Exception (what_arg) { }\
	virtual ~eclass () {;}\
};

//  eclass (const __string& what_arg): logic_error (what_arg) {;}

EXCEPTIONCLASS (generic_exception);
EXCEPTIONCLASS (invalid_format);
EXCEPTIONCLASS (assertion_failed);
EXCEPTIONCLASS (forwarded_failure);
EXCEPTIONCLASS (file_not_found);
EXCEPTIONCLASS (check_failed);

/** If the expr throws an exception, this macro rethrows it, but
 *  records the filename and line number in the Exception reason.
 **/
#define failtrace(expr) try {expr;} catch (Exception e) {throw forwarded_failure (format ("%s\n%%ERR: Exception forwaded at %s[%d]:%s()", (CONSTR) e.what(), __FILE__, __LINE__, __FUNCTION__));}
#define FAILTRACE failtrace

/** Much like failtrace(expr). If the expr throws an exception, this
 *  macro rethrows it, but records the filename and line number in the
 *  Exception reason, as well as additional explanation.
 **/
#define trywith(expr,err) try {expr;} catch (Exception e) {throw forwarded_failure (format ("%s\n%%ERR: Exception forwaded at %s[%d]:%s()\n%%ERR: %s", (CONSTR) e.what(), __FILE__, __LINE__, __FUNCTION__, (CONSTR) err));}
#define TRYWITH trywith

///////////////////////////////////////////////////////////////////////////////
// Debugging macros

/** The @ref Application and ASSERT framework supports many ways of
 *  handling assertion fails. It is, for example, very useful to crash
 *  the program (ASSERT_CRASH) if you are running the program under
 *  debugger.
 *
 *  Just include in your Main line:
 *
 *  assertmode = ASSERT_CRASH;
 **/
enum assertmodes {ASSERT_THROW=0, ASSERT_CRASH=1, ASSERT_EXIT=2, ASSERT_DONTCARE=3};

extern FILE* dbgout;

/** Assertion mode. See @ref assertmodes in magisupp.h
 **/
extern int assertmode;

/** Internal for ASSERT and ASSERTWITH. **/
extern void assertwith (const char* filename, int lineno, const char* funcname,
						const char* condition=NULL, const char* message=NULL);

#ifndef DEBUG_MODE
#define NODEBUG
#endif

#ifndef NODEBUG

#define TRACELINE {fprintf (dbgout, "File:%s, Line#%d %s\n", __FILE__, __LINE__, __FUNCTION__); fflush (dbgout);}
#define TRACE(expr) {fprintf (dbgout, "File:%s, Line#%d %s: ", __FILE__, __LINE__, __FUNCTION__); fprintf (dbgout, expr); fprintf (dbgout, "\n"); fflush (dbgout);}
#define TRACE1(expr,par1) {fprintf (dbgout, "File:%s, Line#%d %s: ", __FILE__, __LINE__, __FUNCTION__); fprintf (dbgout, expr, par1); fprintf (dbgout, "\n"); fflush (dbgout);}
#define TRACE2(expr,par1,par2) {fprintf (dbgout, "File:%s, Line#%d %s: ", __FILE__, __LINE__, __FUNCTION__); fprintf (dbgout, expr, par1, par2); fprintf (dbgout, "\n"); fflush (dbgout);}
#define TRACE3(expr,par1,par2,par3) {fprintf (dbgout, "File:%s, Line#%d %s: ", __FILE__, __LINE__, __FUNCTION__); fprintf (dbgout, expr, par1, par2, par3); fprintf (dbgout, "\n"); fflush (dbgout);}
#define TRACE4(expr,par1,par2,par3,par4) {fprintf (dbgout, "File:%s, Line#%d %s: ", __FILE__, __LINE__, __FUNCTION__); fprintf (dbgout, expr, par1, par2, par3, par4); fprintf (dbgout, "\n"); fflush (dbgout);}

// ASSERT is not available under Qt, which also defines it. Clever.
#ifndef QT_VERSION
#define ASSERT(expr) {if (!(expr)) assertwith (__FILE__, __LINE__, __FUNCTION__, #expr);}
#endif

#define ASSERTWITH(expr,message) {if (!(expr)) assertwith (__FILE__, __LINE__, __FUNCTION__, #expr, message);}
#define PRE(expr) ASSERTWITH(expr, "Violation of precondition");
#define POST(expr) ASSERTWITH(expr, "Violation of postcondition");

#else

#define TRACELINE
#define ASSERT(expr)
#define ASSERTWITH(expr,message)
#define PRE(expr)
#define POST(expr)
#define TRACE(msg) {}
#define TRACE1(msg,p1) {}
#define TRACE2(msg,p1,p2) {}
#define TRACE3(msg,p1,p2,p3) {}
#define TRACE4(msg,p1,p2,p3,p4) {}
#endif

#ifdef MSDOS
#define IOSBINARY binary
#else // UNIX
#define IOSBINARY bin
#endif

// Screen manipulation (minicurses)

#define clrscr() {cout << "\033[H\033[J";}

#define randomize	{srand (time (NULL));}
//#define rnd(range) {return 

EXCEPTIONCLASS (must_overload);
// For abstract classes that can't use =0
#define MUST_OVERLOAD {assertwith (__FILE__, __LINE__, __FUNCTION__, "MUST_OVERLOAD", format("%%ERR: Method of baseclass must be overloaded by descendant %s", typeid(*this).name()));}

EXCEPTIONCLASS (forbidden_call);
#define FORBIDDEN {throw forbidden_call (format("%%ERR: Forbidden execution at %s[%d]: %s()", __FILE__, __LINE__, __FUNCTION__));}

#endif


