///////////////////////////////////////////////////////////////////////////////
//  Copyright Marko Grnroos 1996, all rights reserved.                     //
// Omistusoikeus Marko Grnroos, 1996, kaikki oikeudet pidtetn.           //
//   Tmn tiedoston kopiointi, lukeminen ja kytt ja erityisesti tmn     //
// ilmoitustekstin muuttaminen on kielletty ilman asianmukaista lupaa.      //
//                                                                           //
//                                             Marko Grnroos (magi@utu.fi)  //
///////////////////////////////////////////////////////////////////////////////

#ifndef __MAP_H__
#define __MAP_H__

#include <iostream.h>
#include "magic/object.h"
#include "magic/magisupp.h"
#include "magic/pararr.h"

// Externals
class DumpContext;

class MapPair;
class HashBucket;
class HashFunc;
class GenHash;
class GenHashIter;

/** A key-value pair in @ref GenHash and thus in @ref Map; (internal).
 **/
class MapPair : public Object {
  public:
	const Comparable*	key;
	Object*				value;
	bool				isref;
						MapPair		() {key=NULL; value=NULL; isref=0;}
						~MapPair	();
	int					match		(const Comparable& okey) const {return (*key) == okey;}
	DumpContext&		operator>>	(DumpContext& out) const;
	void				check		() const;

  private:
	decl_dynamic (MapPair);
};

//////////////////////////// Hash-pohjainen mappi /////////////////////////////

/** A hash bucket in @ref GenHash and thus in @ref Map; (internal).
 **/
class HashBucket : public Object {
  public:
	MapPair		pair;
	HashBucket*	next;

					HashBucket	();
					~HashBucket	();
	DumpContext&	operator>>	(DumpContext& out) const;
	void			check		() const;

  private:
	decl_dynamic (HashBucket);
};

/** An interface for hash functions in @ref Genhash (and thus in @ref Map).
 **/
class HashFunc {
  public:
	virtual void	hashfunc	(int hashsize) const = 0;
};

EXCEPTIONCLASS (not_found);

/** Generic hash table storage, used by @ref Map.
 **/
class GenHash : public Object {
  public:

					GenHash		(int hsize=64, int flags=0) {make (NULL, hsize, flags);}
					GenHash		(HashFunc* hfunc, int hsize=16, int flags=0) {
						make (hfunc, hsize, flags);
					}
	void			make		(HashFunc* hashfunc, int hsize, int flags);
	
	void			set			(const Comparable* key, Object* value);
	const Object*	get			(const Comparable& key) const throw (not_found);
	DumpContext&	operator>>	(DumpContext& out) const;
	ostream&		operator>>	(ostream& out) const;
	void			empty		();
	void			remove		(const Comparable& key);
	void			operator+=	(const GenHash& other);

	void			check		() const;
	
  protected:
	Array<HashBucket>	hash;
	int					hashsize;
	HashFunc*			hashfunc;
	bool				isref;

  private:
	decl_dynamic (GenHash);
	friend GenHashIter;
};

/** @ref GenHash iterator.
 **/
class GenHashIter {
  public:

						GenHashIter	(const GenHash* hash);
	
	void				first		();
	void				next		();
	Comparable&			getkeyv		();
	Object&				getvaluev	();
	const Comparable&	getkey		() const;
	const Object&		getvalue	() const;
	int					exhausted	() const {return exh;}

  protected:
	const GenHash*	hash;
	int				bucket;
	HashBucket*		currbucket;
	int				exh;
};

EXCEPTIONCLASS (map_item_not_found);

// Varsinainen kyttliittymluokka.
// Avainluokan on oltava Comparable-perillinen, mutta arvo voi olla mik hyvns

/** Associative map, based on @ref GenHash.
 *
 *  The key class has to inherit @ref Comparable. The value class can
 *  be any inherit of @ref Object.
 *
 *  The inserted objects are owned by the Map by default, and are thus
 *  destroyed along with it, or during various replacement methods. By
 *  giving the MAP_REF flag to the constructor causes the objects not
 *  to be owned by the Map.
 *
 *  Map<String,String> has a shorthand macro StringMap. Many
 *  Map-handling functions, especially for StringMaps, are not
 *  included in the template, but are global. Such methods include
 *  toString(), splitpairs(), and joinpairs(). See section "globals"
 *  for more information, or the header map.h.
 *
 *  Maps can be iterated using @ref MapIter. Note also the forMap and
 *  forStringMap macros in map.h
 **/
template <class keyclass, class valueclass>
class Map : public Object {
  public:
	/** Default constructor uses the default hash size of @ref GenHash.
	 **/
						Map			() {
							hash = new GenHash ();
							isref = 0;
							mFailByThrow=false;
							mFailByThrowOnce=false;
							mFailByNullOnce=false;
						}
	/** Constructor.
	 *
	 *  @param hashsize Hash size, should usually be larger than 16,
	 *  around is typical. Prime numbers are often recommended (which
	 *  16 and 64 are not).
	 *
	 *  @param flags Mode parameters. Currently defined are: MAP_REF
	 *  (causes objects to be not owned by the Map)
	 **/
						Map			(int hashsize, int flags) {
							hash = new GenHash (hashsize, flags);
							isref = flags & MAP_REF;
							mFailByThrow=false;
							mFailByThrowOnce=false;
							mFailByNullOnce=false;
						}
						Map (const Map& orig) {
							hash = new GenHash ();
							isref = 0;
							mFailByThrow=false;
							mFailByThrowOnce=false;
							mFailByNullOnce=false;

							// And copy
							operator= (orig);
						}
						~Map			() {delete hash;}

	/** Sets the _key_ to _value_; copies the given value object using
	 *  the copy constructor.
	 **/
	void				set				(const keyclass& key, const valueclass& value) {
		keyclass* nkey = new keyclass (key);
		valueclass* nvalue = NULL;
		if (isref)
			nvalue = const_cast<valueclass*> (&value);
		else
			nvalue = new valueclass (value);
		hash->set (nkey, nvalue);
	}

	/** Sets the _key_ to _value_; takes ownership of the passed object.
	 **/
	void				set				(const keyclass& key, const valueclass* value) {
		hash->set (new keyclass (key), const_cast<valueclass*> (value));
	}

	/** Removes an item from the map.
	 **/
	void				remove			(const keyclass& key) {
		hash->remove (key);
	}

	/** Queries whether the given key is in the map. */
	bool				hasKey			(const keyclass& key) const {
		return int(hash->get (key));
	}
	
	/** Returns a const reference to object associated to the key.
	 *
	 *  For information about error handling and exceptions, see @ref
	 *  failByThrowOnce().
	 **/
	const valueclass&	operator[]		(const keyclass& key) const {
		return get (key);
	}

	/** Returns a non-const reference to object associated to the key.
	 *
	 *  For information about error handling and exceptions, see @ref
	 *  failByThrowOnce().
	 **/
	valueclass&	operator[]				(const keyclass& key) {
		return const_cast<valueclass&> (get (key));
	}

	/** Returns a const reference to object associated to the key.
	 *
	 *  For information about error handling and exceptions, see @ref
	 *  failByThrowOnce().
	 *
	 *  @see getv
	 **/
	const valueclass&	get				(const keyclass& key) const {
		const valueclass& result = (valueclass&) *hash->get (key);
		if ((mFailByThrow || mFailByThrowOnce) && !mFailByNullOnce) {
			mFailByThrowOnce = false;
			if (isnull(result)) {
				try {
					const keyclass* dummy = &dynamic_cast<const keyclass&> (key);
					dummy=NULL; // Have to do this to avoid a silly warning
				} catch (...) {
					throw map_item_not_found ("Map item not found");
				}
				throw map_item_not_found (format("Map item '%s' not found", (CONSTR) key));
			}
		}
		mFailByNullOnce = false;
		return result;
	}

	/** Returns a non-const reference to object associated to the key.
	 *
	 *  For information about error handling and exceptions, see @ref
	 *  failByThrowOnce().
	 *
	 *  @see get
	 **/
	valueclass&			getv			(const keyclass& key) {
		return const_cast<valueclass&> (get (key));
	}

	/** Returns a const reference to object associated to the key, OR
	 *  if the key is not found from the map, returns the 'def'
	 *  reference back to the caller.
	 *
	 *  For information about error handling and exceptions, see @ref
	 *  failByThrowOnce().
	 *
	 *  @see get
	 **/
	const valueclass&	getOr			(const keyclass& key, const valueclass& def) const {
		valueclass* tmp = (valueclass*) hash->get (key);
		return tmp? *tmp : def;
	}

 	/** Returns a const pointer to object associated to the key, or
	 *  NULL if the object was not found.
	 *
	 *  @see getvp
	 **/
	const valueclass*	getp			(const keyclass& key) const {
		return (valueclass*) hash->get (key);
	}

 	/** Returns a non-const pointer to object associated to the key,
	 *  or NULL if the object was not found.
	 *
	 *  @see getp
	 **/
	valueclass*			getvp			(const keyclass& key) {
		return (valueclass*) hash->get (key);
	}

	// Miscellaneous operators

	/** Implementation for @ref Object. */
	virtual DumpContext&			operator>>	(DumpContext& out) const {
		hash->operator>> (out);
		return out;
	}

	/** Implementation for @ref Object. */
	virtual ostream&			operator>>		(ostream& out) const {
		hash->operator>> (out);
		return out;
	}

	/** Union operator; adds the other Map to self. */
	Map<keyclass,valueclass>&	operator+=	(const Map<keyclass,valueclass>& other) {
		hash->operator+= (*other.hash);
		return *this;
	}

	/** Copy operator. */
	Map<keyclass,valueclass>&	operator=	(const Map<keyclass,valueclass>& other) {
		hash->empty ();
		hash->operator+= (*other.hash);
		isref = other.isref;
		mFailByThrow=other.mFailByThrow;
		mFailByThrowOnce=other.mFailByThrowOnce;
		mFailByNullOnce=other.mFailByNullOnce;
		return *this;
	}

	/** Destroys all the items from the Map.
	 **/
	void				empty			() {
		hash->empty ();
	}

	/** Returns the @ref GenHash of the Map.
	 **/
	const GenHash*		gethash			() const {return hash;}

	/** Implementation for @ref Object. */
	void				check			() const {hash->check();}

	/** Should the access-methods of Map report missing elements by
	 *  throwing an exception (true) or by returning a null reference
	 *  (false, which is the default).
	 *
	 *  Null references to objects can be checked by the global
	 *  isnull() function.
	 **/
	void failByThrow (bool bythrow=true) {mFailByThrow=bythrow;}

	/** Same as @ref failByThrow(true), but applies to only one access
	 *  call.
	 *
	 *  Giving parameter 'false' to this method cancels a previously
	 *  given failByThrowOnce(true) currently in effect.
	 **/
	void				failByThrowOnce	(bool bythrow=true) const {mFailByThrowOnce=bythrow;}

	/** Same as @ref failByThrow(false), but applies to only one
	 *  access call.
	 *
	 *  Giving parameter 'false' to this method cancels a previously
	 *  given failByNullOnce(true) currently in effect.
	 **/
	void				failByNullOnce	(bool bynull=true) const {mFailByNullOnce=bynull;}
	
  private:
	GenHash*		hash;
	bool			isref;
	bool			mFailByThrow;
	mutable bool	mFailByThrowOnce, mFailByNullOnce;
};

enum mapdefs {MAP_NONE=0, MAP_REF=1};

#define CMap Map
typedef Map<String,String> StringMap;


/** Creates a @ref String representation of the Map in format:
 *  {"xxx"="yyy", "zzzz"="aaaa"}
 **/
String toString (const StringMap& map);

/** Tests if the key is in the map. If it is, fetches the value,
 *  otherwise uses the default.
 *  Usage: getOrDefault(myMap, "a_key", 0.1)
 **/
#define getOrDefault(map,key,def)\
	(isnull(map[key])? def : map[key])

/** Constructs the map from string containing key-value-pairs in format:
 *  "key1=value1&key2=value2&key3=value3".
 *
 *  Reverse for @ref joinPairs().
 **/
void splitpairs (Map<String,String>& trg, const String& source,
				 char psep='=', char rsep='&');

/** Constructs a string from the map, representing the key-value-pairs in
 *  format: "key1=value1&key2=value2&key3=value3".
 *
 *  Reverse for @ref splitPairs(). 
 **/
String joinpairs (const Map<String,String>& src, char psep='=', char rsep='&');

/** Map iterator. Usage:
 *
 *  for (MapIter<String,String> iter (myMap); !iter.exhausted(); iter.next())
 *
 *  printf("%s=%s\n", (CONSTR) iter.key(), (CONSTR) iter.value());
 *
 *  , which prints the key-value-pairs of the map to stdout.
 *
 *  See map.h for additional macros forMap and forStringMap for
 *  somewhat easier iteration. For example:
 *
 *  forStringMap (myMap, iter) { printf("%s", (CONSTR) iter.value()); }
 *
 *  , which prints the values in the map to stdout.
 **/
template <class keyclass, class valueclass>
class MapIter {
  public:
	MapIter	(const Map<keyclass,valueclass>& map) : iter (map.gethash()) {;}

	/** Points the iterator to the first item in the @ref Map. */
	void				first		() {iter.first();}

	/** Moves the iterator to next item in the @ref Map. */
	void				next		() {iter.next();}

	/** Returns a const reference to the key in the current position
	 *  of the iterator.
	 **/
	const keyclass&		key			() const {return (const keyclass&) iter.getkey();}

	/** Returns a non-const reference to the key in the current position
	 *  of the iterator.
	 **/
	keyclass&			key			() {return (keyclass&) iter.getkeyv();}

	/** Returns a const reference to the value in the current position
	 *  of the iterator.
	 **/
	const valueclass&	value		() const {return (const valueclass&) iter.getvalue();}

	/** Returns a non-const reference to the value in the current position
	 *  of the iterator.
	 **/
	valueclass&			value		() {return (valueclass&) iter.getvaluev();}

	/** Returns 1 (true) if all the items in the map have been
	 *  iterated, 0 (false) otherwise.
	 **/
	int					exhausted	() const {return iter.exhausted();}

  protected:
	GenHashIter	iter;
};

/** Easier iteration for @ref Map<String,String> alias StringMap. For
 *  example:
 *
 *  forStringMap (myMap, iter) { printf("%s", (CONSTR) iter.value()); }
 *
 *  , which prints the values in the StringMap to stdout.
 **/
#define forStringMap(mapname,varname) \
	for (MapIter<String,String> varname (mapname); !varname.exhausted(); varname.next())

/** Easier iteration for @ref Map. For example:
 *
 *  forMap (String, String, myMap, iter) { printf("%s", (CONSTR) iter.value()); }
 *
 *  , which prints the values in the @ref Map<String,String> to stdout.
 **/
#define forMap(keytype,valuetype,mapname,varname) \
	for (MapIter<keytype,valuetype> varname (mapname); !varname.exhausted(); varname.next())

#endif
