#ifndef __PACKARRAY_H__
#define __PACKARRAY_H__

#include "magic/object.h"
#include "magic/magisupp.h"


//////////////////////////////////////////////////////////////////////////////
//                                                                          //
//        ----              |     _                         /    \          //
//        |   )  ___   ___  |    / \           ___         /      \         //
//        |---   ___| |   \ | / /   \ |/\ |/\  ___| \   | <        >        //
//        |     (   | |     |/  |---| |   |   (   |  \  |  \      /         //
//        |      \__|  \__/ | \ |   | |   |    \__|   \_/   \    /          //
//                                                   \_/                    //
//////////////////////////////////////////////////////////////////////////////

/** Packed array/vector; saves space and is fast. Especially useful
 *  with the primitive types such as int or float. 0-based indexing.
 *
 *  NULL objects are not possible like in @ref Array. Nor is this
 *  container as dynamic as @ref Array is.
 **/
template <class TYPE>
class PackArray : public Object {
  public:
	/** The items as a C-array. */
	TYPE*	data;

	/** Length of the array. */
	int		size;

	PackArray	() {
		data = NULL;
		size = 0;
	}

	/** Creates an packed array with the given size. */
	PackArray	(int siz) {
		data = NULL;
		size = siz;
		make (size);
	}

	/** Copy constructor uses operator= to copy the contained
	 *  items. Remember to define the operator= for the object class,
	 *  as the default operator= does only a shallow copy.
	 **/
	PackArray	(const PackArray& orig) {
		data = NULL;
		make (orig.size);
		for (int i=0; i<size; i++)
			data[i] = orig.data[i];
	}

	/** Creates or recreates the array with the given size. */
	void	make	(int siz) {
		if (data)
			destroy ();
		size = siz;
		if (size>0)
			data = new TYPE [size];
		else
			data = NULL;
	}

	~PackArray	() {
		destroy ();
	}

	/** Very quick copy operator. */
	void	shallowCopy	(const PackArray& orig) {
		if (size != orig.size)
			make (orig.size);
		memcpy (data, orig.data, size*sizeof(TYPE));
	}
	
	/** Empties the array. */
	void	destroy	() {
		delete data;
		data = NULL;
		size = 0;
	}

	/** Empties the array. Alias for @ref destroy(). */
	void	empty () {
		destroy ();
	}

	/** Returns reference to i:th item in the array.
	 *  Const version.
	 **/
	const TYPE&	operator[]	(int i) const {
		ASSERTWITH (i>=0 && i<size,
					format("Index %d out of PackArray (size %d) bounds", i, size));
		return data [i];
	}

	/** Returns reference to i:th item in the array.
	 *  Non-const version.
	 **/
	TYPE&	operator[]	(int i) {
		ASSERTWITH (i>=0 && i<size,
					format("Index %d out of PackArray (size %d) bounds", i, size));
		return data [i];
	}

	/** Standard copy operator.
	 *
	 *  Remember to define the operator= for the object class, as the
	 *  default operator= does only a shallow copy.
	 **/
	operator=	(const PackArray& other) {
		if (this != &other) {
			destroy ();
			if (other.size) {
				data = new TYPE [other.size];
				size = other.size;
				for (int i=0; i<size; i++)	// We have to copy these explicitly
					data[i] = other.data[i];
			}
		}
	}

	/*
	void	removeFill	(int pos) {
		ASSERTWITH (i>=0 && i<size,
					format("Index %d out of PackArray (size %d) bounds", i, size));
		delete data[pos];
		for (int i=pos; i<size-2; i++)
			data[i] = data[i+1];
		data[size-1]=NULL;
		resize (size-1);
	}
	*/
	
	/** Changes size of the array.
	 **/
	void	resize	(int newsize) {
		ASSERT (newsize>=0);
		
		if (newsize == size)
			return;

		if (newsize>0)
			if (data)
				data = (TYPE*) renew (data, sizeof (TYPE)*newsize);
			else
				data = new TYPE [newsize];
		else
			destroy ();

		size = newsize;
	}

	/** Implementation for @ref Object. Archive support. */
	/*
	virtual CArchive&	operator>>	(CArchive& arc) const {
		arc << size;
		for (int i=0; i<size; i++)
			arc << data[i];
		return arc;
	}
	*/
	
	/** Implementation for @ref Object. Archive support. */
	/*
	virtual IStream&	operator<<	(IStream& arc) {
		int nsize;
		arc >> nsize;
		destroy ();
		make (nsize);
		for (int i=0; i<size; i++)
			arc >> data[i];
		return arc;
	}
	*/

	/** Implementation for @ref Object. */
	virtual void	check	() const {
		ASSERT ((data?1:0) == (size?1:0));
		ASSERT (((unsigned int&)*this) != 0xdddddddd);
		if (data)
			ASSERT (((unsigned int&)*data) != 0xdddddddd);
		ASSERT (size>=0);
		ASSERTWITH (size<10000000, "Reasonable array size maximum. Grow if necessary");
	}

  private:
};

#endif
