/***************************************************************************
 *   copyright            : (C) 2000 by Marko Grnroos                     *
 *   email                : magi@iki.fi                                    *
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************
 *
 **/

#ifndef __PARAMETER_H__
#define __PARAMETER_H__

#include<magic/cstring.h>
#include<magic/map.h>


///////////////////////////////////////////////////////////////////////////////
//     ___               ----                                                //
//     |  \          _   |   )  ___       ___         ___   |   ___          //
//     |   | \   | |/ \  |---   ___| |/\  ___| |/|/| /   ) -+- /   ) |/\     //
//     |   |  \  | |   | |     (   | |   (   | | | | |---   |  |---  |       //
//     |__/    \_/ |   | |      \__| |    \__| | | |  \__    \  \__  |       //
//            \_/                                                            //
///////////////////////////////////////////////////////////////////////////////

/** Dynamic field (or object) type parameter.
 *
 **/
class DynParameter {
  public:
	DynParameter () {FORBIDDEN}
	DynParameter (const String& name, const String& shortd, int maxlen, const String& clsname) : mName (name), mShortDescription (shortd), mParamClassName (clsname), mMaxlen(maxlen) {}
	virtual ~DynParameter () {}
		
	const String&	name		() const {return mName;}
	const String&	shortDesc	() const {return mShortDescription;}
	const String&	paramClass	() const {return mParamClassName;}
	int				maxlen		() const {return mMaxlen;}

	/** Checks if the given string representation is a valid
	 *  value. Returns true if it is, false if not.
	 **/
	virtual bool	checkValue	(const String& value) const {return value.len<=mMaxlen;}
	
  private:
	String	mName;
	String	mShortDescription;
	String	mParamClassName;
	int		mMaxlen;
};


/////////////////////////////////////////////////////////////////////////////////
//  ----         o             ----                                            //
// (      |          _         |   )  ___       ___         ___   |   ___      //
//  ---  -+- |/\ | |/ \   ___  |---   ___| |/\  ___| |/|/| /   ) -+- /   ) |/\ //
//     )  |  |   | |   | (   \ |     (   | |   (   | | | | |---   |  |---  |   //
// ___/    \ |   | |   |  ---/ |      \__| |    \__| | | |  \__    \  \__  |   //
//                        __/                                                  //
/////////////////////////////////////////////////////////////////////////////////

class StringParameter : public DynParameter {
  public:
	virtual bool	checkValue	(const String& value) const {return DynParameter::checkValue(value);}
	
  protected:
	StringParameter (const String& name, const String& shortd, int maxlen) : DynParameter (name, shortd, maxlen, "string") {}
};


////////////////////////////////////////////////////////////////////////////////////
// ___                    |       ----                                            //
// |  \             |     |  ___  |   )  ___       ___         ___   |   ___      //
// |   |  __  |   | |---  | /   ) |---   ___| |/\  ___| |/|/| /   ) -+- /   ) |/\ //
// |   | /  \ |   | |   ) | |---  |     (   | |   (   | | | | |---   |  |---  |   //
// |__/  \__/  \__! |__/  |  \__  |      \__| |    \__| | | |  \__    \  \__  |   //
////////////////////////////////////////////////////////////////////////////////////

class DoubleParameter : public DynParameter {
  public:
					DoubleParameter (const String& name, const String& shortd, int maxlen, double minv, double maxv, double defv=-345535.234) : DynParameter (name, shortd, maxlen, "float"), mMin(minv), mMax(maxv), mDefault(defv) {}
	virtual bool	checkValue		(const String& value) const {return DynParameter::checkValue(value);}
	virtual String	defaultString	() const {return (mDefault==-345535.234)? String(""):String(mDefault);}
	
  private:
	double	mMin;
	double	mMax;
	double	mDefault;
};

class IntParameter : public DynParameter {
  public:
					IntParameter	(const String& name, const String& shortd, long minv, long maxv, long defv=-3475235) : DynParameter (name, shortd, 15, "int"), mMin(minv), mMax(maxv), mDefault(defv) {}
	virtual bool	checkValue		(const String& value) const {return DynParameter::checkValue(value);}
	virtual String	defaultString	() const {return (mDefault==-3475235)? String(""):String(mDefault);}
	
  private:
	int	mMin;
	int	mMax;
	int	mDefault;
};

class BoolParameter : public DynParameter {
  public:
					BoolParameter	(const String& name, const String& shortd, int defv=-1) : DynParameter (name, shortd, 1, "int"), mDefault(defv) {}
	virtual bool	checkValue		(const String& value) const {return DynParameter::checkValue(value);}
	virtual String	defaultString	() const {return (mDefault==-1)? String(""):String(mDefault);}
	
  private:
	int		mDefault;
};


///////////////////////////////////////////////////////////////////////////////
//   --- ----                                            o               |   //
//    |  |   )  ___       ___         ___   |   ___        ___  ___      |   //
//    |  |---   ___| |/\  ___| |/|/| /   ) -+- /   ) |/\ |   / /   )  ---|   //
//    |  |     (   | |   (   | | | | |---   |  |---  |   |  /  |---  (   |   //
//   _|_ |      \__| |    \__| | | |  \__    \  \__  |   | /__  \__   ---|   //
///////////////////////////////////////////////////////////////////////////////

/** Interface for dynamically parameterizable objects.
 *
 **/
class IParameterized {
  public:
	/** Returns description of parameters required by the training
	 *  algorithm.
	 **/
	virtual Array<DynParameter>*	parameters	() const {MUST_OVERLOAD; return NULL;}

	/** Initializes the object with given dynamic parameters.
	 **/
	virtual void					init		(const StringMap& params) {MUST_OVERLOAD}
};

/** The INITPARAMS macro handles and prints errors in IParameterized::init() function.
 *
 */
#define INITPARAMS(mapname, initcode) \
	try { const_cast<StringMap&>(mapname).failByThrow(); initcode \
	} catch (Exception e) {\
		String errorstr = e.what();\
		errorstr += "\nInitialization parameter missing. Parameters are:\n";\
		Array<DynParameter>* pars = parameters ();\
		for (int i=0; i<pars->size; i++)\
			errorstr += (*pars)[i].name () + " ";\
		errorstr += '\n';\
		throw Exception (errorstr);\
	}

#endif
