#ifndef __TABLE_H__
#define __TABLE_H__

#include<magic/object.h>
#include<magic/cstring.h>
#include<magic/magisupp.h>

//////////////////////////////////////////////////////////////////////////////
//                                                                          //
//        ----              |   -----             |         /    \          //
//        |   )  ___   ___  |     |    ___  |     |  ___   /      \         //
//        |---   ___| |   \ | /   |    ___| |---  | /   ) <        >        //
//        |     (   | |     |/    |   (   | |   ) | |---   \      /         //
//        |      \__|  \__/ | \   |    \__| |__/  |  \__    \    /          //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

/** Packed two-dimensional table; works also with native types.
 *
 *  Ideal for storing numeric matrices; any mathematical matrix
 *  implementation would most probably inherit PackTable<double>.
 **/
template <class TYPE>
class PackTable : public Object {
  protected:
	TYPE*	data;
  public:
	/** Creates a null-sized table. Dimensions can be given later with
	 *  the @ref make() method.
	 **/
	PackTable	() {
		data = NULL;
		rows = 0,
		cols = 0;
	}

	/** Creates a Table.
	 *
	 *  @param rs Number of rows in the Table.
	 *  @param cs Number of columns in the Table.
	 **/
	PackTable	(int rs, int cs) {
		data = NULL;
		rows = 0,
		cols = 0;
		make (rs, cs);
	}

	/** Copy constructor. */
	PackTable	(const PackTable& orig) {
		data = NULL;
		rows = 0,
		cols = 0;
		make (orig.rows, orig.cols);
		for (int i=0; i<rows; i++)
			for (int j=0; j<cols; j++)
				get(i,j) = orig.get(i,j);
	}
	
	~PackTable	() {
		destroy ();
	}

	/** Destroys the content of the Table, as well as its
	 *  dimensions. Can be resized later with the @ref make() method.
	 **/
	void	destroy	() {
		delete data;
		data = NULL;
		rows = 0,
		cols = 0;
	}

	/** Creates or recreates the Table with the given dimensions. If
	 *  the Table has any old contents, they are destroyed.
	 *
	 *  @param nrows Number of rows in the Table.
	 *  @param ncols Number of columns in the Table.
	 **/
	virtual void	make	(int nrows, int ncols) {
		ASSERT (nrows>=0 && ncols>=0);
		if (data)
			destroy ();
		rows = nrows,
		cols = ncols;
		if (rows>0 && cols>0) {
			// Luodaan uusi tietorakenne valmiiksi
			newComment (format ("PackTable<> data, %d rows, %d cols", rows, cols));
			data = new TYPE [rows*cols];
		}
	}

	/** Retrieves a reference to an item in the Table. Const
	 *  version.
	 *
	 *  @param row Row index of the item. 0-based.
	 *  @param col Column index of the item. 0-based.
	 **/
	const TYPE&	get	(int row, int col) const {
		ASSERTWITH (row>=0 && row<rows && col>=0 && col<cols,
					format ("Table range overflow: %d,%d out of %d,%d",
							row, col, rows, cols));
		return data[row*cols+col];
	}

	/** Retrieves a reference to an item in the Table. Non-const
	 *  version.
	 *
	 *  @param row Row index of the item. 0-based.
	 *  @param col Column index of the item. 0-based.
	 **/
	TYPE&	get	(int row, int col) {
		ASSERTWITH (row>=0 && row<rows && col>=0 && col<cols,
					format ("Table range overflow: %d,%d out of %d,%d",
							row, col, rows, cols));
		return data[row*cols+col];
	}

	/** Serialization implementation for @ref Object.
	 **/
	/*
	virtual CArchive&	operator>>	(CArchive& arc) const {
		arc << rows << cols;
		for (int i=0; i<rows; i++)
			for (int j=0; j<cols; j++)
				arc << get (i, j);
		return arc;
	}
	*/

	/** Serialization implementation for @ref Object.
	 **/
	/*
	IStream&	operator<<	(IStream& arc) {
		int nrows, ncols;
		arc >> nrows >> ncols;
		make (nrows, ncols);
		for (int i=0; i<rows; i++)
			for (int j=0; j<cols; j++)
				arc >> get (i, j);
		return arc;
	}
	*/

	/** Number of rows in the table. */
	int	rows;

	/** Number of columns in the table. */
	int	cols;
	
  private:
	operator=	(const PackTable& other) {}
};

#endif
